#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <assert.h>
#include <math.h>


const double epsilon = 1e-6;


//  ____  _      _                  
// |  _ \(_) ___| |_ _   _ _ __ ___ 
// | |_) | |/ __| __| | | | '__/ _ \
// |  __/| | (__| |_| |_| | | |  __/
// |_|   |_|\___|\__|\__,_|_|  \___|
//                                  
#define PIXEL_W 1024
#define PIXEL_H 1024
// Create picture
uint8_t* pict_alloc() {
  return (uint8_t*)malloc(3 * PIXEL_W * PIXEL_H);
}
// Set pixel in picture
void set(uint8_t* pict, unsigned x, unsigned y, uint8_t r, uint8_t g, uint8_t b) {
  assert(x < PIXEL_W);
  assert(y < PIXEL_H);

  pict[3 * y * PIXEL_W + 3 * x] = b;
  pict[3 * y * PIXEL_W + 3 * x + 1] = g;
  pict[3 * y * PIXEL_W + 3 * x + 2] = r;
}
// Get pixel colors
void get(uint8_t* pict, unsigned x, unsigned y, uint8_t* r, uint8_t* g, uint8_t* b) {
  assert(x < PIXEL_W);
  assert(y < PIXEL_H);

  *r = pict[3 * y * PIXEL_W + 3 * x];
  *g = pict[3 * y * PIXEL_W + 3 * x + 1];
  *b = pict[3 * y * PIXEL_W + 3 * x + 2];
}

void write_picture_to_file(const char* path, uint8_t* pict) {
  FILE* f = fopen(path, "w");
  unsigned filesize = 54 + 3*PIXEL_W*PIXEL_H;

  // Header
  unsigned char bmpfileheader[14] = {'B','M', 0,0,0,0, 0,0, 0,0, 54,0,0,0};
  bmpfileheader[ 2] = (unsigned char)(filesize    );
  bmpfileheader[ 3] = (unsigned char)(filesize>> 8);
  bmpfileheader[ 4] = (unsigned char)(filesize>>16);
  bmpfileheader[ 5] = (unsigned char)(filesize>>24);
  fwrite(bmpfileheader, 1, 14, f);
  
  // Info
  unsigned char bmpinfoheader[40] = {40,0,0,0, 0,0,0,0, 0,0,0,0, 1,0, 24,0};
  bmpinfoheader[ 4] = (unsigned char)( PIXEL_W    );
  bmpinfoheader[ 5] = (unsigned char)( PIXEL_W>> 8);
  bmpinfoheader[ 6] = (unsigned char)( PIXEL_W>>16);
  bmpinfoheader[ 7] = (unsigned char)( PIXEL_W>>24);
  bmpinfoheader[ 8] = (unsigned char)( PIXEL_H    );
  bmpinfoheader[ 9] = (unsigned char)( PIXEL_H>> 8);
  bmpinfoheader[10] = (unsigned char)( PIXEL_H>>16);
  bmpinfoheader[11] = (unsigned char)( PIXEL_H>>24);
  fwrite(bmpinfoheader, 1, 40, f);

  // Pixels
  unsigned char bmppad[3] = {0,0,0};
  for(size_t y = 0; y < PIXEL_H; ++y) {
    fwrite(pict + 3 * (PIXEL_H - y - 1) * PIXEL_W, 3, PIXEL_W, f);
    fwrite(bmppad, 1, (4 - (PIXEL_W*3)%4)%4, f);
  }

  fclose(f);
}


//   ____                      _           
//  / ___|___  _ __ ___  _ __ | | _____  __
// | |   / _ \| '_ ` _ \| '_ \| |/ _ \ \/ /
// | |__| (_) | | | | | | |_) | |  __/>  < 
//  \____\___/|_| |_| |_| .__/|_|\___/_/\_\
//                      |_|                
struct C {
  double real, im;
};
typedef struct C C;
C add(C x, C y) {
  C ret;
  ret.real = x.real + y.real;
  ret.im = x.im + y.im;
  return ret;
}
C mult(C x, C y) {
  C ret;
  ret.real = x.real * y.real - x.im * y.im;
  ret.im = x.im * y.real + x.real * y.im;
  return ret;
}
double norm(C x) {
  return sqrt(x.real * x.real + x.im * x.im);
}
double theta(C x) {
  return atan2(x.real, x.im);
}
// Give the pixel coordinate of a points, given the top left and bottom right complex points
void clip(C top_left, C bottom_right, C c, unsigned* x, unsigned* y) {
  double w = bottom_right.real - top_left.real;
  double h = bottom_right.im - top_left.im;
  double cx = (c.real - top_left.real) * (double)PIXEL_W / w;
  double cy = (c.im - top_left.im) * (double)PIXEL_H / h;
  *x = cx;
  *y = cy;
}




//  __  __       _       
// |  \/  | __ _(_)_ __  
// | |\/| |/ _` | | '_ \ 
// | |  | | (_| | | | | |
// |_|  |_|\__,_|_|_| |_|
//                       
int main() {
  uint8_t* picture = pict_alloc();
  assert(picture != NULL);
  for(size_t x = 0; x < PIXEL_W; ++x) {
    for(size_t y = 0; y < PIXEL_H; ++y) {
      uint8_t r = 255.0 * (sin(5.0 * M_PI * (double)x / (double)PIXEL_W) + 1.0) / 2.0;
      uint8_t g = 255.0 * (sin(10.0 * M_PI * (double)y / (double)PIXEL_H) + 1.0) / 2.0;
      uint8_t b = 255.0 * (sin(3.0 * M_PI * (double)(x+y) / (double)PIXEL_H) + 1.0) / 2.0;
      set(picture, x, y, r, g, b);
    }
  }

  write_picture_to_file("test.bmp", picture);

  free(picture);
  return 0;
}
